# Rain Droplet Community Credit Software
# Copyright (C) 2006-11
# 
# This program is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the 
# Free Software Foundation; either version 2 of the License, or (at your 
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program; if not, write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Based on RippleSite by Ryan Fugger
# ryan@ripplepay.com
#-------------------------------------------------------------------------

from django.core import meta

#todo: validators?
#todo: add admin tags, default values, etc. - nice model things
#todo: split Node up into User and Node with 1-1 relationship to speed up node fetching??

DIGITS = 24 # max. number of digits in amounts
DECIMALS = 12 # digits reserved for decimals

class CurrencyUnit(meta.Model):
  short_name = meta.CharField(maxlength=5, unique=True)
  long_name = meta.CharField(maxlength=50)
  symbol = meta.CharField(maxlength=3)
  value = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS, default=0.0) # in base units - for conversion
  last_update = meta.DateTimeField(auto_now=True) # last value update

  def __repr__(self):
    return self.short_name

  class META:
    admin = meta.Admin(list_display=('long_name', 'short_name', 'symbol', 'value', 'last_update'))
    ordering = ('short_name',)

class Node(meta.Model):
  username = meta.CharField(maxlength=20, unique=True)
  name = meta.CharField(maxlength=50)
  pwd = meta.CharField(maxlength=40)
  salt = meta.CharField(maxlength=10, default='') # for salting the password before hashing
  display_units = meta.ForeignKey(CurrencyUnit, related_name='node_default', null=True, blank=True, default=None)
  do_conversions = meta.BooleanField(default=True) # allow conversion on through payments
  #notify_emails = meta.BooleanField()
  #bulletins = meta.BooleanField()
  location = meta.FloatField(max_digits=12, decimal_places=11) # routing location

  def __repr__(self):
    return self.name or self.username
  
  def checkPwd(self, pwd):
    import sha
    if self.pwd == sha.new(pwd + self.salt).hexdigest():
      return True
    else: return False
  
  def setPwd(self, pwd):
    import sha
    self.pwd = sha.new(pwd + self.salt).hexdigest()
  
  def getPrimaryEmail(self):
    return self.get_email_list(primary__exact=True)[0].email
  
  class META:
    admin = meta.Admin(list_display=('username', 'name', 'display_units', 'do_conversions'),
                       list_filter=('display_units', 'do_conversions'),
                       search_fields=('username', 'name'))

class EmailAddr(meta.Model):
  node = meta.ForeignKey(Node, related_name='email', db_index=True)
  email = meta.EmailField(unique=True)
  code = meta.CharField(maxlength=10, unique=True)
  confirmed = meta.BooleanField(default=False)
  primary = meta.BooleanField()
  
  def __repr__(self):
    return self.email

  class META:
    admin = meta.Admin(list_display=('email', 'node', 'confirmed', 'primary'),
                       list_filter=('confirmed', 'primary'),
                       search_fields=('email',))

class Reminder(meta.Model):
  node = meta.ForeignKey(Node, related_name='reminder')
  code = meta.CharField(maxlength=10, unique=True)
  date = meta.DateTimeField(auto_now_add=True)
  
  def __repr__(self):
    return "%s %s" % (self.node, self.date)

  class META:
    admin = meta.Admin(list_display=('node', 'code', 'date'))

# accounts ----------------------------------
class SharedAccountData(meta.Model):
  currency = meta.ForeignKey(CurrencyUnit)
  balance = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)
  active = meta.BooleanField(default=True)
  created = meta.DateTimeField(auto_now_add=True)
  
  # interest
  interest_rate = meta.FloatField(max_digits=13, decimal_places=10, default=0.0) # stored as decimal for faster computation (as opposed to percentage for display)
  proposed_rate = meta.FloatField(max_digits=13, decimal_places=10, null=True, blank=True, default=None)
  node_to_confirm = meta.ForeignKey(Node, related_name='shared_account_to_confirm', null=True, blank=True, default=None)
  last_update = meta.DateTimeField(auto_now_add=True) # to quickly compute present interest tally during payments
  # (interest is added to balance each update, ie, each transaction)

  def displayRate(self):
    return self.interest_rate * 100.0
  def displayProposedRate(self):
    if self.proposed_rate <> None:
      return self.proposed_rate * 100.0
    return None
  
  def __repr__(self):
    str = '%s account %d' % (self.get_currency(), self.id)
    if not self.active:
      str += " (inactive)"
    return str
  
  class META:
    admin = meta.Admin(list_display=('currency', 'balance', 'active', 'interest_rate', 'last_update'),
                       list_filter=('currency', 'active'))
  
  # avoid rounding on save when django converts to string for SQL command
  def _pre_save(self):
    if type(self.balance) == type(3.14):
      self.balance = '%.12f' % self.balance
  
  def _post_save(self):
    if type(self.balance) == type('3.14'):
      self.balance = float(self.balance)

class Account(meta.Model):
  owner = meta.ForeignKey(Node, related_name='account', db_index=True)
  partner = meta.ForeignKey(Node, related_name='partner_account', db_index=True)
  shared_data = meta.ForeignKey(SharedAccountData)
  balance_multiplier = meta.IntegerField(choices=[(1, '+1'), ( -1, '-1')])
  partner_acct = meta.ForeignKey("self", null=True, blank=True)

  # what owner calls the account (default would be name/email of partner)
  name = meta.CharField(maxlength=50)

  # current max negative (debt) balance of this account
  iou_limit = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)

  # meta-limits - acceptable range for upper and lower account limits according to owner
  # used to allow each partner to regulate their own credit flexibly
  # iou_limit = min(my_limit, partner_acct.partner_limit)
  partner_limit = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS) # max amt to accept from partner
  my_limit = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS, null=True, blank=True) # max amt of debt to owe to partner - null -> no limit

  def __repr__(self):
    return '%d->%d (%d: "%s")' % (self.owner_id, self.partner_id, self.shared_data_id, self.name)

  def getAvailableCredit(self):
    return self.iou_limit + self.getBalance() #- frozen

  def getBalance(self):
    return self.getBalanceNoInterest() + self.getPresentInterest()

  def getBalanceNoInterest(self):
    return self.get_shared_data().balance * self.balance_multiplier

  def getPresentInterest(self):
    from raindroplet.droplet.views.pathgraph import computeInterest
    from datetime import datetime
    sharedData = self.get_shared_data()
    p = sharedData.balance * self.balance_multiplier
    return computeInterest(p, sharedData.interest_rate, sharedData.last_update, datetime.now())

  class META:
    admin = meta.Admin(list_display=('id', 'name', 'owner', 'shared_data', 'iou_limit', 'partner_limit', 'my_limit'),
                       search_fields=('name', 'owner__name', 'owner__username', 'partner__name', 'partner__username'))

class Offer(meta.Model):
  initiator = meta.ForeignKey(Node, related_name='sent_offer')
  recipient_email = meta.CharField(maxlength=50)
  amount = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)
  currency = meta.ForeignKey(CurrencyUnit)
  interest_rate = meta.FloatField(max_digits=13, decimal_places=8) # stored as percentage, 2 fewer decimals

  def __repr__(self):
    return "%s offer %s %.2f to %s" % (self.get_initiator(), self.get_currency(), self.amount, self.recipient_email)

  class META:
    admin = meta.Admin(list_display=('initiator', 'recipient_email', 'amount', 'currency'),
                       list_filter=('currency',),
                       search_fields=('initiator__name', 'initiator__username', 'recipient_email'))


# payments ---------------------------
PAYMENT_STATUS_CHOICES = (
  ('PE', 'Pending'),
  ('OK', 'Completed'),
  ('CA', 'Cancelled'),
  ('FA', 'Failed'),
)

class Payment(meta.Model):
  payer = meta.ForeignKey(Node, related_name='outgoing_payment', db_index=True)
  payer_email = meta.EmailField()
  recipient = meta.ForeignKey(Node, related_name='incoming_payment', db_index=True)
  recipient_email = meta.EmailField()
  date = meta.DateTimeField(auto_now_add=True)
  currency = meta.ForeignKey(CurrencyUnit)
  amount = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)
  status = meta.CharField(maxlength=2, choices=PAYMENT_STATUS_CHOICES)
  description = meta.TextField(default='', blank=True)

  def __repr__(self):
    return "%d: %s pay %s%.2f to %s" % (self.id, self.get_payer(), self.get_currency().symbol, self.amount, self.get_recipient())

  class META:
    admin = meta.Admin(list_display=('id', 'payer', 'recipient', 'date', 'currency', 'amount', 'status'),
                       list_filter=('currency', 'status'),
                       search_fields=('payer__name', 'payer__username', 'payer_email', 'recipient__name', 'recipient__username', 'recipient_email', 'description'))
  
  # avoid rounding on save when django converts to string for SQL command
  def _pre_save(self):
    if type(self.amount) == type(3.14):
      self.amount = '%.12f' % self.amount
  
  def _post_save(self):
    if type(self.amount) == type('3.14'):
      self.amount = float(self.amount)
    

# payment may have multiple paths
class PaymentPath(meta.Model):
  payment = meta.ForeignKey(Payment, related_name='path')
  amount = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)

  def __repr__(self):
    return "%d: %.2f on payment %s" % (self.id, self.amount, self.get_payment())

  #class META:
  #  admin = meta.Admin(list_display=('id', 'payment', 'amount'))
  
  # avoid rounding on save when django converts to string for SQL command
  def _pre_save(self):
    if type(self.amount) == type(3.14):
      self.amount = '%.12f' % self.amount
  
  def _post_save(self):
    if type(self.amount) == type('3.14'):
      self.amount = float(self.amount)


# each link is an account from the perspective of the partner closest to payment recipient
# (the IOU recipient for this link)
class PaymentLink(meta.Model):
  path = meta.ForeignKey(PaymentPath, related_name='link')
  payer_account = meta.ForeignKey(Account)
  position = meta.PositiveIntegerField() # position in path sequence, starting at 1

  # needed because this account may be in a different currency than payment
  amount = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS)
  interest = meta.FloatField(max_digits=DIGITS, decimal_places=DECIMALS) # add this to amount to get amount actually subtracted from acct balance during transaction
  interest_rate = meta.FloatField(max_digits=13, decimal_places=10) # for auditability

  def __repr__(self):
    return "%d. %s on path %s" % (self.position, self.get_payer_account(), self.get_path())

  #class META:
  #  admin = meta.Admin(list_display=('path', 'payer_account', 'position', 'amount'))
  
  # avoid rounding on save when django converts to string for SQL command
  def _pre_save(self):
    if type(self.amount) == type(3.14):
      self.amount = '%.12f' % self.amount
    if type(self.interest) == type(3.14):
      self.interest = '%.12f' % self.interest
  
  def _post_save(self):
    if type(self.amount) == type('3.14'):
      self.amount = float(self.amount)
    if type(self.interest) == type('3.14'):
      self.interest = float(self.interest)

