"""
Main views file
"""

from django.core import template_loader
from django.core.extensions import DjangoContext
from django.utils.httpwrappers import HttpResponse, HttpResponseRedirect
from django.core.meta import Q
from mail import send_mail
from django.utils.html import escape

from django.models.droplet import nodes, emailaddrs, sharedaccountdatas, accounts, currencyunits, offers, payments, reminders
import routing
import django.conf.settings as settings

import datetime, re, random, sha
from smtplib import SMTPException
from socket import gaierror


def home(request):
  userNode = checkLogin(request)
  if not userNode: # not logged in - show home page
    request.session.set_test_cookie()

    d = {}
    # some usage data for front page
    d['num_users'] = nodes.get_count()
    d['num_accounts'] = sharedaccountdatas.get_count()
    d['num_payments'] = payments.get_count()
    
    t = template_loader.get_template('home')
    c = DjangoContext(request, d)
    return HttpResponse(t.render(c))
  else: # logged in - show Summary
    return HttpResponseRedirect('/summary/')

def summary(request):
  """The users home page"""
  d = {}

  # get user's Node
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  d['userNode'] = userNode

  # get any error messages from session
  d['infos'] = getSessionInfos(request)

  # outstanding offers of credit received/made
  d['receivedOffers'] = offers.get_list(recipient_email__in=getEmails(userNode), order_by=('-id',))
  #d['sentOffers'] = userNode.get_sent_offer_list(order_by=('-id',))

  d['rateProposals'] = getRateProposalAccts(userNode)
  
  # number of connections
  d['numAccounts'] = userNode.get_account_count(shared_data__active__exact=True)

  # account balances etc. summary
  acctUnits = getAccountsSummary(userNode)
  d['totalBalance'] = 0.0
  d['totalOutCredit'] = 0.0
  d['totalInCredit'] = 0.0
  for unit in acctUnits:
    unit.display = True
    unit.displayBalance = unit.balance
    unit.displayOutCredit = unit.availOutCredit
    unit.displayInCredit = unit.availInCredit
    # if displaying in default units, convert and sum
    if userNode.display_units_id:
      if unit.value:
        unit.displayBalance *= unit.value / userNode.get_display_units().value
        d['totalBalance'] += unit.displayBalance
        unit.displayOutCredit *= unit.value / userNode.get_display_units().value
        d['totalOutCredit'] += unit.displayOutCredit
        unit.displayInCredit *= unit.value / userNode.get_display_units().value
        d['totalInCredit'] += unit.displayInCredit
      else: # can't convert to default units, so can't display
        unit.display = False

  convertibleUnits = currencyunits.get_list(value__gt=0.0, order_by=('long_name',))
  hasConvertibleAcct = False
  for unit in acctUnits:
    if unit in convertibleUnits:
      hasConvertibleAcct = True
      break
  if not hasConvertibleAcct:
    convertibleUnits = None
  #else: # move default units to the top of the list
  #  for unit in convertibleUnits:
  #    if unit.id == userNode.display_units_id:
  #      convertibleUnits[:0] = [convertibleUnits.pop(convertibleUnits.index(unit))]
  d['accountUnits'] = acctUnits
  d['convertibleUnits'] = convertibleUnits

  # get recent payments list
  from payment import prepPaymentsForDisplay # do it here to avoid circular import problem
  d['payments'] = payments.get_list(
    complex=Q(payer__pk=userNode.id) | (Q(recipient__pk=userNode.id)), status__exact='OK',
    order_by=('-date',),
    limit=4)
  prepPaymentsForDisplay(d['payments'], userNode)

  t = template_loader.get_template('summary')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

# retrieves user's node if logged in - otherwise returns None
def checkLogin(request):
  try: # reload node from id every time to be safe, since it may have been changed by last request
    nodeId = request.session['userNodeId']
    userNode = nodes.get_object(pk=nodeId)
  except KeyError, NodeDoesNotExist:
    return None
  return userNode

# get 'infos' key from session - used to pass informational messages for the user between pages
def getSessionInfos(request):
  if 'infos' in request.session:
    infos = request.session['infos']
    del request.session['infos']
  else:
    infos = []
  return infos

""" takes a node and returns a list of currencies, each with the following variables hacked in:
 - accts (node's accounts denominated in that currency)
 - balance (cumulative balance of accounts in that currency)
 - lowerLimit (as a positive number - actual balance limit is the negative of this)
 - upperLimit
 - availOutCredit (= balance + lowerLimit)
 - availInCredit (= upperLimit - balance)
"""
def getAccountsSummary(node):
  units = []
  for acct in node.get_account_list(shared_data__active__exact=True, order_by=('name',)):
    acct.balance = acct.getBalance() # only compute present interest once
    unit = acct.get_shared_data().get_currency()
    if unit not in units:
      unit.accts = []
      unit.balance = 0.0
      unit.lowerLimit = 0.0
      unit.availOutCredit = 0.0
      unit.upperLimit = 0.0
      unit.availInCredit = 0.0
      units.append(unit)
    unit = units[units.index(unit)] # load existing one to keep extra variables
    unit.accts.append(acct)
    unit.balance += acct.balance
    unit.lowerLimit += acct.iou_limit
    unit.availOutCredit += max(acct.balance + acct.iou_limit, 0.0)
    unit.upperLimit += acct.get_partner_acct().iou_limit
    unit.availInCredit += max(acct.get_partner_acct().iou_limit - acct.balance, 0.0)

  units.sort(lambda a, b: cmp(a.long_name, b.long_name))
  return units

def login(request):
  """POST: login user, GET: display login form"""
  userNode = checkLogin(request)
  if userNode:
    return HttpResponseRedirect('/')
  
  d = {}
  errors = []
  if request.POST:
    if request.session.test_cookie_worked():
      request.session.delete_test_cookie()
    else: # cookie error requires a separate error page so user can enable cookies and reload login page to set test cookie again.
      t = template_loader.get_template('error')
      c = DjangoContext(request, {'message':'Please enable cookies and try your login again.', 'link':'/login/'})
      return HttpResponse(t.render(c))
    redirect = request.POST['redirect']
    username = request.POST['username']
    d['username'] = username
    password = request.POST['password']
    email = None
    
    if nodes.get_count(username__iexact=username) > 0: # user entered username
      userNode = nodes.get_object(username__iexact=username)
      email = emailaddrs.get_object(node__pk=userNode.id, primary__exact=True)
    
    elif emailaddrs.get_count(email__iexact=username) > 0: # user entered email address
      email = emailaddrs.get_object(email__iexact=username)
      userNode = email.get_node()
    
    # check password
    if email and userNode.checkPwd(password):
      if email.confirmed: # login
          request.session['userNodeId'] = email.node_id # store only the id and reload every request in case of change
          return HttpResponseRedirect(redirect)
      else:
        errors.append("You have not yet confirmed email address '%s' by clicking the link in your confirmation email.  Please login using your old email address or username." % escape(email.email))
        if sendConfirmationEmail(email.email, email.code, request):
          errors.append('Your confirmation email has been resent.  If you do not receive it, <a href="/contact/">please contact us.</a>')
        else:
          errors.append('Your confirmation email could not be resent. <a href="/contact/">Please contact us</a> if you cannot find your original confirmation email.')
    else:
      errors.append('Invalid login details. Please try again or <a href="/register/">click here to sign up as a new user</a>.')

  # display login form
  d['infos'] = errors
  if request.GET:
    d['redirect'] = request.GET['redirect']
  request.session.set_test_cookie()
  t = template_loader.get_template('login')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def logout(request):
  if 'userNodeId' in request.session:
    del request.session['userNodeId']
  return HttpResponseRedirect('/')

def register(request):
  """POST: register new user, GET: display registration form"""
  errors = []
  if request.POST:
    if request.session.test_cookie_worked():
      request.session.delete_test_cookie()
    else: # cookie error requires a separate error page so user can enable cookies and reload login page to set test cookie again.
      t = template_loader.get_template('error')
      c = DjangoContext(request, {'message':'Please enable cookies and try your registration again.', 'link':'/register/'})
      return HttpResponse(t.render(c))

    username = request.POST['username']
    if nodes.get_count(username__iexact=username) > 0:
      errors.append("Username '%s' is taken." % username)

    name = request.POST['name']
    if name == '': name = username # default to username if name left blank

    email = request.POST['email']
    if emailaddrs.get_count(email__iexact=email) > 0:
      errors.append("Someone has already registered email '%s'." % escape(email))

    password = request.POST['password']
    confirmPassword = request.POST['confirmPassword']
    if password != confirmPassword:
      errors.append('The two passwords entered do not match.')
    if len(password) < 4:
      errors.append('Your password must be at least four characters long.')

    displayUnitsStr = request.POST['displayUnits']
    if displayUnitsStr == 'None':
      displayUnits = None
    else: # load units
      displayUnits = currencyunits.get_object(short_name__exact=displayUnitsStr)

    if username == '' or email == '' or password == '':
      errors.append('You must fill in every field with a star beside it.')
      
    if not errors:
      # send email confirmation first
      code = makeConfirmationCode()
      if sendConfirmationEmail(email, code, request):
        hashedPwd = hash(password + code)
        newUser = nodes.Node(username=username, name=name, salt=code[:10], pwd=hashedPwd, location=routing.getNewLocation(), display_units=displayUnits, do_conversions=True)
        newUser.save()
        newEmail = emailaddrs.EmailAddr(node=newUser, email=email, code=code, confirmed=False, primary=True)
        newEmail.save()
        # redirect to informational page
        return HttpResponseRedirect('/registrationSuccess/')
      else:
        errors.append('User not registered. A confirmation email could not be sent to %s. \
            Please verfiy email address and/or try again later.' % escape(email))
    
    # add general error message
    generalError = 'There were errors with your registration.  Please correct them and try again.'
    errors = [generalError] + errors
  
  # display registration form
  request.session.set_test_cookie()
  d = {}
  d['infos'] = errors
  d['units'] = currencyunits.get_list(value__gt=0.0, order_by=('long_name',)) # must be convertible to be default units
  if request.POST:
    d['username'] = request.POST['username']
    d['name'] = request.POST['name']
    d['email'] = request.POST['email']
    d['displayUnits'] = request.POST['displayUnits']
  t = template_loader.get_template('registration')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def makeConfirmationCode():
  while True:
    code = str(random.randint(1000000000, 9999999999))
    if emailaddrs.get_count(code__exact=code) == 0:
      break
  return code

def hash(s):
  return sha.new(s).hexdigest()

# returns True if email sent, False otherwise
def sendConfirmationEmail(email, code, request):
  t = template_loader.get_template('emailConfirmation')
  c = DjangoContext(request, {'email':email, 'code':code})
  return sendEmail('Please confirm email address', t.render(c), email)

def sendEmail(subject, msg, recipient, sender=None, reply_to=None, attempts=3, includeServiceName=True):
  if sender == None:
    sender = '"%s" <%s>' % settings.ADMINS[0]
  if includeServiceName:
    subject = '%s - %s' % (settings.SERVICE_NAME, subject)
  while attempts > 0:
    try:
      if send_mail(subject, msg, sender, (recipient,), reply_to) == 1:
      	return True
    except (SMTPException, gaierror):
      attempts -= 1
  return False

def registrationSuccess(request):
  t = template_loader.get_template('registrationSuccess')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def confirmEmail(request, code):
  userNode = checkLogin(request)
  if emailaddrs.get_count(code__exact=code) > 0:
    email = emailaddrs.get_object(code__exact=code)
    if email.confirmed:
      t = template_loader.get_template('error')
      c = DjangoContext(request, {'message':'That email address has already been confirmed.', 'link':'/login/'})
      return HttpResponse(t.render(c))
  else:
    t = template_loader.get_template('error')
    c = DjangoContext(request, {'message':'Invalid confirmation code.', 'link':'/'})
    return HttpResponse(t.render(c))
  
  # set email to 'confirmed'/primary, and others for same node to secondary
  email.confirmed = True
  email.save()
  
  if userNode:
    if email.get_node() == userNode:
      request.session['infos'] = ["Email address %s confirmed." % escape(email.email)]
      if emailaddrs.get_count(node__pk=userNode.id) == 1:
        return HttpResponseRedirect('/summary/') # first time, send to summary
      else: 
        return HttpResponseRedirect('/profile/')
    else: # logged in as different user
      # logout
      if 'userNodeId' in request.session:
        del request.session['userNodeId']
  
  request.session.set_test_cookie() # necessary because this is a login page
  t = template_loader.get_template('confirmEmail')
  c = DjangoContext(request, {'username':email.email})
  return HttpResponse(t.render(c))

def profile(request):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  d = {}
  
  # get any error messages from session
  d['infos'] = getSessionInfos(request)

  if request.POST:
    name = request.POST['name']
    if name != '' and name != userNode.name:
      userNode.name = name
      d['infos'].append("Name changed to %s." % name)
    
    password = request.POST['password']
    confirmPassword = request.POST['confirmPassword']
    if password != confirmPassword:
      d['infos'].append('The two new passwords entered do not match.  Password not changed.')
    elif len(password) < 4:
      d['infos'].append('Passwords must be at least four characters long.  Password not changed.')
    elif password != '':
      userNode.pwd = hash(password + userNode.salt)
      d['infos'].append('Password changed.')
    
    displayUnits = request.POST['displayUnits']
    if displayUnits == 'None' and userNode.display_units_id:
      userNode.display_units_id = None
      d['infos'].append('Account display units set to None.')
    elif currencyunits.get_count(short_name__exact=displayUnits) > 0:
      curr = currencyunits.get_object(short_name__exact=displayUnits)
      if userNode.display_units_id != curr.id:
        userNode.display_units_id = curr.id
        d['infos'].append('Default account display units changed to %s.' % escape(displayUnits))
    
    if request.POST['convert'] == 'no':
      doConversions = False
    else:
      doConversions = True
    if doConversions != userNode.do_conversions:
      userNode.do_conversions = doConversions
      if doConversions:
        d['infos'].append('Conversions between account units enabled.')
      else:
        d['infos'].append('Conversions between account units disabled.')
    
    userNode.save()
  
  d['userNode'] = userNode
  d['units'] = currencyunits.get_list(value__gt=0.0, order_by=('long_name',)) # must be convertible to be default units
  if userNode.display_units_id:
    d['displayUnits'] = userNode.get_display_units()
  else:
    d['displayUnits'] = None
  d['emails'] = userNode.get_email_list(order_by=('-primary',))
  t = template_loader.get_template('profile')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

# takes string and returns a list of email addresses in the string with no repeats
def parseEmails(text):
  patternLenient = r"\w+([-+.]\w+)*@\w+([-.]\w+)*\.\w+([-.]\w+)*" # lenient email-matching pattern
  matches = re.finditer(patternLenient, text)
  emails = []
  for match in matches:
    nextEmail = text[match.start():match.end()]
    if nextEmail not in emails:
      emails.append(nextEmail)
  return list(emails)

def isEmail(email):
  emails = parseEmails(email)
  if len(emails) == 1 and emails[0] == email:
    return True
  else: return False

# context processor for automatically inserting userNode into DjangoContext
# activated in settings.py under TEMPLATE_CONTEXT_PROCESSORS
def userNode_proc(request):
  return {'userNode':checkLogin(request)}
  
# context processor for automatically inserting site info into DjangoContext
# activated in settings.py under TEMPLATE_CONTEXT_PROCESSORS
def siteInfo_proc(request):
  d = {}
  d['SITE_NAME'] = settings.SITE_NAME
  d['SERVICE_NAME'] = settings.SERVICE_NAME
  d['ADMIN_NAME'] = settings.ADMINS[0][0]
  d['ADMIN_EMAIL'] = settings.ADMINS[0][1]
  return d

# used to show IE 6 and below a GIF instead of the nice PNG logo
def userAgent_proc(request):
  agent = request.META['HTTP_USER_AGENT']
  ie = (agent.count('MSIE') > 0 and agent.count('MSIE 7') == 0)
  return {'iexplorer': ie}

# generic class for holding any types of data
class Struct:
  pass

def construction(request):
  t = template_loader.get_template('construction')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def contact(request):
  t = template_loader.get_template('contact')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def about(request):
  t = template_loader.get_template('about')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def faq(request):
  t = template_loader.get_template('faq')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def getEmails(node, confirmed=True):
  if confirmed:
    emails = emailaddrs.get_list(node__pk=node.id, confirmed__exact=True)
  else:
    emails = emailaddrs.get_list(node__pk=node.id)
  return [e.email for e in emails]

def emailAction(request, emailId, action):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=/profile/')
  
  if emailaddrs.get_count(pk=emailId) == 0:
    return HttpResponseRedirect('/profile/')
  
  email = emailaddrs.get_object(pk=emailId)
  if email.node_id != userNode.id:
    return HttpResponseRedirect('/profile/')
  
  if action == 'delete':
    if not email.primary:
      email.delete()
      request.session['infos'] = ["Email address %s deleted from profile." % escape(email.email)]
    else:
      request.session['infos'] = ["Cannot delete primary email address."]
  
  elif action == 'makePrimary':
    if email.confirmed:
      oldPrimary = emailaddrs.get_object(node__pk=userNode.id, primary__exact=True)
      oldPrimary.primary = False
      oldPrimary.save()
      email.primary = True
      email.save()
      request.session['infos'] = ["%s is now your primary email." % escape(email.email)]
    else:
      request.session['infos'] = ["Cannot make %s your primary address until it is confirmed." % escape(email.email)]
  
  elif action == 'sendConfirmation':
    if not email.confirmed:
      if sendConfirmationEmail(email.email, email.code, request):
        request.session['infos'] = ["Confirmation email resent to %s." % escape(email.email)]
      else:
        request.session['infos'] = ['Confirmation email could not be resent to %s.  Please try again in a few minutes or <a href="/contact/">contact us.</a>' % escape(email.email)]
    else:
      request.session['infos'] = ["%s already confirmed." % escape(email.email)]
      
  return HttpResponseRedirect('/profile/')

def addEmail(request):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  if not request.POST:
    return HttpResponseRedirect('/profile/')
  
  email = request.POST['addEmail']
  if not isEmail(email):
    request.session['infos'] = ["'%s' is not a valid email address.  Please try again." % escape(email)]
    return HttpResponseRedirect('/profile/')
  
  if emailaddrs.get_count(email__exact=email) > 0:
    request.session['infos'] = ["%s is already registered." % escape(email)]
    return HttpResponseRedirect('/profile/')
  
  code = makeConfirmationCode()
  if not sendConfirmationEmail(email, code, request):
    request.session['infos'] = ['Confirmation email could not be resent to %s.  Please try again in a few minutes or <a href="/contact/">contact us.</a>' % escape(email.email)]
    return HttpResponseRedirect('/profile/')
  
  newEmail = emailaddrs.EmailAddr(node=userNode, email=email, code=code, confirmed=False, primary=False)
  newEmail.save()
  request.session['infos'] = ["A confirmation email has been sent to %s." % escape(email)]
    
  return HttpResponseRedirect('/profile/')

def essay(request):
  t = template_loader.get_template('essay')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def forgotPassword(request):
  errors = []
  if request.POST:
    email = request.POST['email']
    if not isEmail(email):
      errors.append("%s does not appear to be a valid email address.  Please try again." % escape(email))
    
    if not errors:
      if emailaddrs.get_count(email__exact=email) > 0:
        emailAddr = emailaddrs.get_object(email__exact=email)
        code = makeReminderCode()
        t = template_loader.get_template('emailForgotPassword')
        c = DjangoContext(request, {'email':email, 'code':code})
        if sendEmail('How to reset your password', t.render(c), email):
          node = emailAddr.get_node()
          existingReminders = node.get_reminder_list()
          for reminder in existingReminders:
            reminder.delete() # only one oustanding reminder per node allowed
          newReminder = reminders.Reminder(node=emailAddr.get_node(), code=code)
          newReminder.save()
        else:
          errors.append("Could not send email to %s.  Please verify the email address and try again." % escape(email))
      else: # email address not in system
        pass # pretend like we're sending a reminder anyways, so no one knows whether an email is registered or not
    
    if not errors:
      return HttpResponseRedirect('/passwordReminderSent/')
  
  d = {}
  d['infos'] = errors
  t = template_loader.get_template('forgotPassword')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def makeReminderCode():
  while True:
    code = str(random.randint(1000000000, 9999999999))
    if reminders.get_count(code__exact=code) == 0:
      break
  return code

def passwordReminderSent(request):
  t = template_loader.get_template('passwordReminderSent')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def resetPassword(request, code):
  if reminders.get_count(code__exact=code):
    reminder = reminders.get_object(code__exact=code)
  else:
    return HttpResponseRedirect('/')
  
  errors = []
  if request.POST: # change password
  
    # *** add security question check here
    
    pwd = request.POST['password']
    confirm = request.POST['confirm']
    if pwd != confirm:
      errors.append("The two new passwords do not match.  Please try again.")
    elif len(pwd) < 4:
      errors.append("Passwords must have at least four characters.  Please try again.")
    if not errors:
      node = reminder.get_node()
      node.pwd = hash(pwd + node.salt)
      node.save()
      reminder.delete()
      return HttpResponseRedirect('/resetPasswordSuccess/')
  
  # display reset password form
  d = {}
  d['infos'] = errors
  d['code'] = code
  t = template_loader.get_template('resetPassword')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))
  
def resetPasswordSuccess(request):
  t = template_loader.get_template('resetPasswordSuccess')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def donate(request):
  t = template_loader.get_template('donate')
  c = DjangoContext(request, {})
  return HttpResponse(t.render(c))

def getRateProposalAccts(node):
  return accounts.get_list(owner__pk=node.id, shared_data__node_to_confirm__pk=node.id)

